import logging

from contracts_lib_py.utils import process_fulfill_condition

from nevermined_sdk_py.nevermined.keeper import NeverminedKeeper as Keeper

logger = logging.getLogger(__name__)


def fulfill_lock_payment_condition(event, agreement_id, price, consumer_account, lock_condition_id):
    """
    Fulfill the lock reward condition.

    :param event: AttributeDict with the event data.
    :param agreement_id: id of the agreement, hex str
    :param price: Asset price, int
    :param consumer_account: Account instance of the consumer
    :param lock_condition_id: hex str the id of the lock reward condition for this `agreement_id`
    """
    if not event:
        logger.warning(
            f'`fulfill_lock_payment_condition` got empty event: event listener timed out.')
        return

    keeper = Keeper.get_instance()
    if keeper.condition_manager.get_condition_state(lock_condition_id) > 1:
        logger.debug(f'lock payment condition already fulfilled/aborted: '
                     f'agreementId={agreement_id}, lockPayment conditionId={lock_condition_id}')
        return

    logger.debug(f"about to lock payment (agreement {agreement_id}) after event {event}.")

    approved = keeper.token.token_approve(
        keeper.lock_payment_condition.address, price, consumer_account)
    logger.info(f'approval of token transfer was {"" if approved else "NOT"} successful')
    args = (
        agreement_id,
        keeper.escrow_payment_condition.address,
        price,
        consumer_account
    )
    process_fulfill_condition(args, keeper.lock_payment_condition, lock_condition_id, logger, keeper,
                              10)


fulfillLockRewardCondition = fulfill_lock_payment_condition



from common_utils_py.agreements.service_agreement import ServiceAgreement
from common_utils_py.agreements.service_types import ServiceTypes, ServiceTypesIndices
from common_utils_py.utils.utilities import to_checksum_addresses
from web3 import providers

from nevermined_sdk_py.gateway.gateway import Gateway
from nevermined_sdk_py.nevermined.agreements import check_token_address
from nevermined_sdk_py.nevermined.keeper import NeverminedKeeper as Keeper
from tests.resources.helper_functions import get_buyer_public_key, get_key, get_provider_babyjub_key, get_provider_public_key, log_event
from tests.resources.mocks.gateway_mock import GatewayMock
from tests.resources.snark_util import call_prover

def test_sign_agreement(publisher_instance, consumer_instance, proof_ddo):
    # point consumer_instance's Gateway mock to the publisher's nevermined instance
    Gateway.set_http_client(
        GatewayMock(publisher_instance, publisher_instance.main_account))

    consumer_acc = consumer_instance.main_account
    keeper = Keeper.get_instance()

    publisher_acc = publisher_instance.main_account

    did = proof_ddo.did
    asset_id = proof_ddo.asset_id
    ddo = consumer_instance.assets.resolve(did)
    service_agreement = ServiceAgreement.from_ddo(ServiceTypes.ASSET_ACCESS_PROOF, ddo)

    price = service_agreement.get_price()

    # Give consumer some tokens
    keeper.dispenser.request_vodkas(price * 2, consumer_acc)
    consumer_acc.babyjub_address = get_buyer_public_key()
    publisher_acc.babyjub_address = get_provider_public_key()
    provider_secret = get_provider_babyjub_key().secret
    proof = call_prover(consumer_acc.babyjub_address, provider_secret, '0x'+get_key())

    agreement_id, signature = consumer_instance.agreements.prepare(
        did, consumer_acc, ServiceTypesIndices.DEFAULT_ACCESS_PROOF_INDEX)

    success = publisher_instance.agreements.create(
        did,
        ServiceTypesIndices.DEFAULT_ACCESS_PROOF_INDEX,
        agreement_id,
        consumer_acc,
        publisher_acc
    )
    assert success, 'createAgreement failed.'

    # Verify condition types (condition contracts)
    agreement_values = keeper.agreement_manager.get_agreement(agreement_id)
    assert agreement_values.did == asset_id, ''
    cond_types = keeper.access_proof_template.get_condition_types()
    for i, cond_id in enumerate(agreement_values.condition_ids):
        cond = keeper.condition_manager.get_condition(cond_id)
        assert cond.type_ref == cond_types[i]

    access_cond_id, lock_cond_id, escrow_cond_id = agreement_values.condition_ids

    # Fulfill lock_payment_condition is done automatically when create agreement is done correctly
    assert 2 == keeper.condition_manager.get_condition_state(lock_cond_id), ''
    assert 1 == keeper.condition_manager.get_condition_state(access_cond_id), ''
    assert 1 == keeper.condition_manager.get_condition_state(escrow_cond_id), ''

    # Fulfill access_condition
    tx_hash = keeper.access_proof_condition.fulfill(
        agreement_id, proof['hash'], consumer_acc.babyjub_address, publisher_acc.babyjub_address, proof['cipher'], proof['proof'], publisher_acc
    )
    keeper.access_proof_condition.get_tx_receipt(tx_hash)
    assert 2 == keeper.condition_manager.get_condition_state(access_cond_id), ''
    event = keeper.access_proof_condition.subscribe_condition_fulfilled(
        agreement_id,
        10,
        log_event(keeper.access_proof_condition.FULFILLED_EVENT),
        (),
        wait=True
    )
    assert event, 'no event for AccessSecretStoreCondition.Fulfilled'

    # Fulfill escrow_payment_condition

    amounts = service_agreement.get_amounts_int()
    receivers = service_agreement.get_receivers()
    token_address = check_token_address(
        keeper, service_agreement.get_param_value_by_name('_tokenAddress'))

    tx_hash = keeper.escrow_payment_condition.fulfill(
        agreement_id, asset_id, amounts, receivers,
        keeper.escrow_payment_condition.address, token_address, lock_cond_id,
        access_cond_id, publisher_acc
    )
    keeper.escrow_payment_condition.get_tx_receipt(tx_hash)
    assert 2 == keeper.condition_manager.get_condition_state(escrow_cond_id), ''
    event = keeper.escrow_payment_condition.subscribe_condition_fulfilled(
        agreement_id,
        10,
        log_event(keeper.escrow_payment_condition.FULFILLED_EVENT),
        (),
        wait=True
    )
    assert event, 'no event for EscrowReward.Fulfilled'
    publisher_instance.assets.retire(did)


def test_agreement_status(setup_agreements_proof_environment, agreements):
    (
        keeper,
        publisher_acc,
        consumer_acc,
        agreement_id,
        asset_id,
        price,
        service_agreement,
        (lock_cond_id, access_cond_id, escrow_cond_id),

    ) = setup_agreements_proof_environment

    consumer_acc.babyjub_address = get_buyer_public_key()
    publisher_acc.babyjub_address = get_provider_public_key()
    provider_secret = get_provider_babyjub_key().secret
    proof = call_prover(consumer_acc.babyjub_address, provider_secret, '0x'+get_key())

    success = keeper.access_proof_template.create_agreement(
        agreement_id,
        asset_id,
        [access_cond_id, lock_cond_id, escrow_cond_id],
        service_agreement.conditions_timelocks,
        service_agreement.conditions_timeouts,
        consumer_acc.address,
        publisher_acc
    )
    assert success, f'createAgreement failed {success}'
    event = keeper.access_proof_template.subscribe_agreement_created(
        agreement_id,
        10,
        log_event(keeper.access_proof_template.AGREEMENT_CREATED_EVENT),
        (),
        wait=True
    )
    assert event, 'no event for AgreementCreated '
    assert agreements.status(agreement_id) == {"agreementId": agreement_id,
                                               "conditions": {"lockReward": 1,
                                                              "accessProof": 1,
                                                              "escrowReward": 1
                                                              }
                                               }

    amounts = service_agreement.get_amounts_int()
    receivers = service_agreement.get_receivers()

    token_address = keeper.token.address

    agreements.conditions.lock_payment(
        agreement_id, asset_id, amounts, receivers, token_address, consumer_acc)

    event = keeper.lock_payment_condition.subscribe_condition_fulfilled(
        agreement_id,
        10,
        log_event(keeper.lock_payment_condition.FULFILLED_EVENT),
        (),
        wait=True
    )
    assert event, 'no event for LockRewardCondition.Fulfilled'
    assert agreements.status(agreement_id) == {"agreementId": agreement_id,
                                               "conditions": {"lockReward": 2,
                                                              "accessProof": 1,
                                                              "escrowReward": 1
                                                              }
                                               }
    tx_hash = keeper.access_proof_condition.fulfill(
        agreement_id, proof['hash'], consumer_acc.babyjub_address, publisher_acc.babyjub_address, proof['cipher'], proof['proof'], publisher_acc
    )
    keeper.access_proof_condition.get_tx_receipt(tx_hash)
    event = keeper.access_proof_condition.subscribe_condition_fulfilled(
        agreement_id,
        20,
        log_event(keeper.access_proof_condition.FULFILLED_EVENT),
        (),
        wait=True
    )
    assert event, 'no event for AccessProofCondition.Fulfilled'
    assert agreements.status(agreement_id) == {"agreementId": agreement_id,
                                               "conditions": {"lockReward": 2,
                                                              "accessProof": 2,
                                                              "escrowReward": 1
                                                              }
                                               }

    tx_hash = keeper.escrow_payment_condition.fulfill(
        agreement_id, asset_id, amounts, receivers,
        keeper.escrow_payment_condition.address, token_address, lock_cond_id,
        access_cond_id, publisher_acc
    )
    keeper.escrow_payment_condition.get_tx_receipt(tx_hash)
    event = keeper.escrow_payment_condition.subscribe_condition_fulfilled(
        agreement_id,
        10,
        log_event(keeper.escrow_payment_condition.FULFILLED_EVENT),
        (),
        wait=True
    )
    assert event, 'no event for EscrowReward.Fulfilled'
    assert agreements.status(agreement_id) == {"agreementId": agreement_id,
                                               "conditions": {"lockReward": 2,
                                                              "accessProof": 2,
                                                              "escrowReward": 2
                                                              }
                                               }



from tests.resources.snark_util import poseidon_hash
from nevermined_sdk_py.config_provider import ConfigProvider
from examples.example_config import ExampleConfig
import json
import logging
import logging.config
import os
import pathlib
import uuid
from urllib.request import urlopen

import coloredlogs
import yaml
from common_utils_py.ddo.ddo import DDO
from common_utils_py.http_requests.requests_session import get_requests_session
from contracts_lib_py.utils import get_account
from common_utils_py.agreements.service_factory import ServiceDescriptor
from nevermined_sdk_py.gateway.gateway_provider import GatewayProvider
from nevermined_sdk_py.gateway.gateway import Gateway
from nevermined_sdk_py.secret_store.secret_store import SecretStore
from nevermined_sdk_py.nevermined.keeper import NeverminedKeeper as Keeper
from nevermined_sdk_py.nevermined.nevermined import Nevermined
from nevermined_sdk_py.secret_store.secret_store_provider import SecretStoreProvider
from tests.resources.mocks.gateway_mock import GatewayMock
from tests.resources.mocks.secret_store_mock import SecretStoreMock
from common_utils_py.utils.utilities import to_checksum_addresses

PUBLISHER_INDEX = 1
CONSUMER_INDEX = 0


def get_resource_path(dir_name, file_name):
    base = os.path.realpath(__file__).split(os.path.sep)[1:-1]
    if dir_name:
        return pathlib.Path(os.path.join(os.path.sep, *base, dir_name, file_name))
    else:
        return pathlib.Path(os.path.join(os.path.sep, *base, file_name))


def init_ocn_tokens(nevermined, account, amount=100):
    nevermined.accounts.request_tokens(account, amount)
    Keeper.get_instance().token.token_approve(
        Keeper.get_instance().dispenser.address,
        amount,
        account
    )


def get_publisher_account():
    return get_account(0)


def get_consumer_account():
    return get_account(1)


def get_publisher_instance(init_tokens=True, use_ss_mock=True, use_gateway_mock=True):
    config = ExampleConfig.get_config()
    ConfigProvider.set_config(config)
    nevermined = Nevermined()
    account = get_publisher_account()
    nevermined.main_account = account
    if init_tokens:
        init_ocn_tokens(nevermined, nevermined.main_account)
    if use_ss_mock:
        SecretStoreProvider.set_secret_store_class(SecretStoreMock)
    else:
        SecretStoreProvider.set_secret_store_class(SecretStore)
    if use_gateway_mock:
        GatewayProvider.set_gateway_class(GatewayMock)
    else:
        Gateway.set_http_client(get_requests_session())
        GatewayProvider.set_gateway_class(Gateway)

    return nevermined


def get_consumer_instance(init_tokens=True, use_ss_mock=True, use_gateway_mock=True):
    config = ExampleConfig.get_config()
    ConfigProvider.set_config(config)
    nevermined = Nevermined()
    account = get_consumer_account()
    nevermined.main_account = account
    if init_tokens:
        init_ocn_tokens(nevermined, nevermined.main_account)
    if use_ss_mock:
        SecretStoreProvider.set_secret_store_class(SecretStoreMock)
    else:
        SecretStoreProvider.set_secret_store_class(SecretStore)
    if use_gateway_mock:
        GatewayProvider.set_gateway_class(GatewayMock)
    else:
        Gateway.set_http_client(get_requests_session())
        GatewayProvider.set_gateway_class(Gateway)

    return nevermined


def _get_asset(url):
    return DDO(json_text=get_assset_json_text(url))


def get_assset_json_text(url):
    return json.dumps(json.loads(urlopen(url).read().decode('utf-8')))


def get_ddo_sample():
    return _get_asset(
        "https://raw.githubusercontent.com/nevermined-io/docs/master/docs/architecture/specs/examples/metadata/v0.1/ddo-example-access.json")


def get_algorithm_ddo():
    return json.loads(urlopen(
        "https://raw.githubusercontent.com/nevermined-io/docs/1a853613bd9e9d633474e571d3225378a37b6451/docs/architecture/specs/examples/metadata/v0.1/ddo-example-algorithm.json").read().decode(
        'utf-8'))


def get_workflow_ddo():
    return json.loads(urlopen(
        "https://raw.githubusercontent.com/nevermined-io/docs/master/docs/architecture/specs/examples/metadata/v0.1/ddo-example-workflow.json").read().decode(
        'utf-8'))


def get_computing_metadata():
    return json.loads(urlopen(
        'https://raw.githubusercontent.com/nevermined-io/docs/master/docs/architecture/specs/examples/metadata/v0.1/computing-metadata.json').read().decode(
        'utf-8'))


def get_sample_nft_ddo():
    return json.loads(urlopen(
        'https://raw.githubusercontent.com/keyko-io/nevermined-docs/master/docs/architecture/specs'
        '/examples/access/v0.1/ddo_nft.json').read().decode(
        'utf-8'))


def get_computing_ddo():
    return _get_asset(
        "https://raw.githubusercontent.com/nevermined-io/docs/master/docs/architecture/specs/examples/compute/v0.1/ddo.computing.json")


def get_registered_ddo(nevermined_instance, account):
    metadata = get_metadata()
    asset_rewards = {
        "_amounts": ["10", "2"],
        "_receivers": ["0x00Bd138aBD70e2F00903268F3Db08f2D25677C9e", "0x068ed00cf0441e4829d9784fcbe7b9e26d4bd8d0"]
    }
    metadata['main']['files'][0]['checksum'] = str(uuid.uuid4())
    ddo = nevermined_instance.assets.create(metadata, account, asset_rewards=asset_rewards)
    return ddo

def get_key():
    return "4e657665726d696e65640a436f707972696768742032303230204b65796b6f20476d62482e0a0a546869732070726f6475637420696e636c75646573"

class BabyjubKey:
    def __init__(self, secret, public1, public2):
        self.secret = secret
        self.x = public1
        self.y = public2

def get_provider_babyjub_key():
    secret = os.getenv('PROVIDER_BABYJUB_SECRET', '')
    public1 = os.getenv('PROVIDER_BABYJUB_PUBLIC1', '')
    public2 = os.getenv('PROVIDER_BABYJUB_PUBLIC2', '')
    return BabyjubKey(secret, public1, public2)

def get_provider_public_key():
    public1 = os.getenv('PROVIDER_BABYJUB_PUBLIC1', '')
    public2 = os.getenv('PROVIDER_BABYJUB_PUBLIC2', '')
    return [public1, public2]

def get_buyer_public_key():
    public1 = os.getenv('BUYER_BABYJUB_PUBLIC1', '')
    public2 = os.getenv('BUYER_BABYJUB_PUBLIC2', '')
    return [public1, public2]

def get_proof_ddo(nevermined_instance, account):
    metadata = get_metadata()
    asset_rewards = {
        "_amounts": ["10", "2"],
        "_receivers": ["0x00Bd138aBD70e2F00903268F3Db08f2D25677C9e", "0x068ed00cf0441e4829d9784fcbe7b9e26d4bd8d0"]
    }
    key=get_key()
    metadata['main']['files'][0]['url'] = key
    metadata['main']['files'][0]['checksum'] = str(uuid.uuid4())
    hash = poseidon_hash(key)
    providerKey = get_provider_public_key()
    metadata['additionalInformation'] = {
        "providerKey": {
            "x": providerKey[0],
            "y": providerKey[1]
        },
        "poseidonHash": hash
    }
    access_service_attributes = {
        "main": {
            "name": "dataAssetAccessProofServiceAgreement",
            "creator": account.address,
            "price": metadata['main']['price'],
            "timeout": 3600,
            "datePublished": metadata['main']['dateCreated'],
            "_amounts": asset_rewards['_amounts'],
            "_hash": hash,
            "_providerPub": providerKey,
            "_receivers": asset_rewards['_receivers']
        }
    }
    service_descriptor = ServiceDescriptor.access_proof_service_descriptor(
        access_service_attributes,
        'http://localhost:8030'
    )
    ddo = nevermined_instance.assets.create(metadata, account, [service_descriptor], asset_rewards=asset_rewards)
    return ddo


def get_registered_ddo_nft(nevermined_instance, account):
    ddo = get_sample_nft_ddo()
    metadata = ddo['service'][0]['attributes']
    metadata['main']['files'][0][
        'url'] = "https://raw.githubusercontent.com/tbertinmahieux/MSongsDB/master/Tasks_Demos" \
                 "/CoverSongs/shs_dataset_test.txt?" + str(uuid.uuid4())
    metadata['main']['files'][0]['checksum'] = str(uuid.uuid4())

    escrow_payment_condition = ddo['service'][1]['attributes']['serviceAgreementTemplate']['conditions'][2]
    _number_nfts = 1
    _amounts = get_param_value_by_name(escrow_payment_condition['parameters'], '_amounts')
    _receivers = to_checksum_addresses(
        get_param_value_by_name(escrow_payment_condition['parameters'], '_receivers'))

    transfer_nft_condition = ddo['service'][1]['attributes']['serviceAgreementTemplate']['conditions'][1]
    _nftHolder = get_param_value_by_name(transfer_nft_condition['parameters'], '_nftHolder')


    _total_price = 0
    for i in _amounts:
        _total_price += int(i)

    asset_rewards = {
        "_amounts": _amounts,
        "_receivers": _receivers
    }

    metadata['main']['price'] = _total_price

    access_service_attributes = {"main": {
        "name": "nftAccessAgreement",
        "creator": account.address,
        "timeout": 3600,
        "price": _total_price,
        "_amounts": _amounts,
        "_receivers": _receivers,
        "_numberNfts": str(_number_nfts),
        "datePublished": metadata['main']['dateCreated']
    }}

    sales_service_attributes = access_service_attributes
    sales_service_attributes['main']['name'] = 'nftSalesAgreement'
    sales_service_attributes['main']['_nftHolder'] = _nftHolder

    nft_sales_service_descriptor = ServiceDescriptor.nft_sales_service_descriptor(
        sales_service_attributes,
        'http://localhost:8030'
    )
    nft_access_service_descriptor = ServiceDescriptor.nft_access_service_descriptor(
        access_service_attributes,
        'http://localhost:8030'
    )
    ddo = nevermined_instance.assets.create(metadata, account,
                                            [nft_sales_service_descriptor, nft_access_service_descriptor],
                                            asset_rewards=asset_rewards,
                                            royalties= 0, cap=10, mint=10)
    return ddo


def get_registered_with_psk(nevermined_instance, account, auth_method):
    metadata = get_ddo_sample()
    ddo = nevermined_instance.assets.create(metadata.metadata, account,
                                            authorization_type=auth_method)
    return ddo


def log_event(event_name):
    def _process_event(event):
        print(f'Received event {event_name}: {event}')

    return _process_event


def get_rsa_private_key_file():
    return os.getenv('RSA_PRIVKEY_FILE', '')


def get_rsa_public_key_file():
    return os.getenv('RSA_PUBKEY_FILE', '')


def get_metadata():
    return json.loads(urlopen(
        "https://raw.githubusercontent.com/nevermined-io/docs/master/docs/architecture/specs"
        "/examples/metadata/v0.1/metadata1.json").read().decode(
        'utf-8'))


def setup_logging(default_path='logging.yaml', default_level=logging.INFO, env_key='LOG_CFG'):
    """Logging setup."""
    path = default_path
    value = os.getenv(env_key, None)
    if value:
        path = value
    if os.path.exists(path):
        with open(path, 'rt') as file:
            try:
                config = yaml.safe_load(file.read())
                logging.config.dictConfig(config)
                coloredlogs.install()
                logging.info(f'Logging configuration loaded from file: {path}')
            except Exception as ex:
                print(ex)
                print('Error in Logging Configuration. Using default configs')
                logging.basicConfig(level=default_level)
                coloredlogs.install(level=default_level)
    else:
        logging.basicConfig(level=default_level)
        coloredlogs.install(level=default_level)


def get_param_value_by_name(parameters, name):
    """
    Return the value from the conditions parameters given the param name.

    :return: Object
    """
    for p in parameters:
        if p['name'] == name:
            return p['value']


from common_utils_py.agreements.service_factory import ServiceDescriptor


class Services:
    """Nevermined services class."""

    @staticmethod
    def create_access_service(attributes, service_endpoint):
        """
        Publish an asset with an `Access` service according to the supplied attributes.

        :param attributes: attributes of the access service, dict
        :param service_endpoint: str URL for initiating service access request
        :return: Service instance or None
        """
        service = ServiceDescriptor.access_service_descriptor(attributes, service_endpoint)
        return service



import uuid
from unittest.mock import Mock, MagicMock

import pytest
from common_utils_py.agreements.service_agreement import ServiceAgreement, ServiceAgreementTemplate
from common_utils_py.agreements.service_types import ServiceTypes
from common_utils_py.did import DID
from common_utils_py.metadata.metadata import Metadata
from common_utils_py.utils.utilities import generate_prefixed_id
from contracts_lib_py.contract_handler import ContractHandler
from contracts_lib_py.web3_provider import Web3Provider
from contracts_lib_py.web3.http_provider import CustomHTTPProvider
from web3 import Web3

from examples import ExampleConfig
from nevermined_sdk_py import ConfigProvider
from nevermined_sdk_py.nevermined.keeper import NeverminedKeeper as Keeper
from nevermined_sdk_py.assets.asset_executor import AssetExecutor
from nevermined_sdk_py.assets.asset_consumer import AssetConsumer
from nevermined_sdk_py.nevermined.agreements import Agreements
from tests.resources.helper_functions import (_get_asset, get_algorithm_ddo, get_buyer_public_key, get_consumer_account,
                                              get_consumer_instance, get_ddo_sample,
                                              get_metadata, get_proof_ddo, get_provider_public_key, get_publisher_account,
                                              get_publisher_instance, get_registered_ddo, get_workflow_ddo,
                                              setup_logging)
from tests.resources.mocks.secret_store_mock import SecretStoreMock

setup_logging()

ConfigProvider.set_config(ExampleConfig.get_config())


@pytest.fixture(autouse=True)
def setup_all():
    config = ExampleConfig.get_config()
    Web3Provider._web3 = Web3(CustomHTTPProvider(config.keeper_url))
    ContractHandler.artifacts_path = config.keeper_path
    Keeper.get_instance(artifacts_path=config.keeper_path)


@pytest.fixture
def metadata_provider_instance():
    config = ExampleConfig.get_config()
    ConfigProvider.set_config(config)
    return Metadata(ConfigProvider.get_config().metadata_url)

@pytest.fixture
def secret_store():
    return SecretStoreMock


@pytest.fixture
def publisher_instance():
    return get_publisher_instance()


@pytest.fixture
def consumer_instance():
    return get_consumer_instance()


@pytest.fixture
def publisher_instance_no_init():
    return get_publisher_instance(False, False, False)


@pytest.fixture
def consumer_instance_no_init():
    return get_consumer_instance(False, False, False)


@pytest.fixture
def publisher_instance_gateway():
    return get_publisher_instance(use_gateway_mock=False, use_ss_mock=False)


@pytest.fixture
def consumer_instance_gateway():
    return get_consumer_instance(use_gateway_mock=False, use_ss_mock=False)


@pytest.fixture
def registered_ddo():
    return get_registered_ddo(get_publisher_instance(), get_publisher_account())

@pytest.fixture
def proof_ddo():
    return get_proof_ddo(get_publisher_instance(), get_publisher_account())


@pytest.fixture
def web3_instance():
    config = ExampleConfig.get_config()
    return Web3Provider.get_web3(config.keeper_url)


@pytest.fixture
def asset1():
    asset = _get_asset(
        'https://raw.githubusercontent.com/nevermined-io/docs/master/docs/architecture/specs'
        '/examples/access/v0.1/ddo1.json')
    asset._did = DID.encoded_did(asset.proof['checksum'])
    return asset

@pytest.fixture
def asset2():
    asset = _get_asset(
        'https://raw.githubusercontent.com/nevermined-io/docs/master/docs/architecture/specs'
        '/examples/access/v0.1/ddo2-update.json')
    asset._did = DID.encoded_did(asset.proof['checksum'])
    return  asset

@pytest.fixture
def ddo_sample():
    ddo = get_ddo_sample()
    ddo.metadata['main']['files'][0]['checksum'] = str(uuid.uuid4())
    return ddo


@pytest.fixture
def ddo_algorithm():
    ddo = get_algorithm_ddo()
    ddo['service'][0]['attributes']['main']['files'][0]['checksum'] = str(uuid.uuid4())
    return ddo


@pytest.fixture
def ddo_workflow():
    ddo = get_workflow_ddo()
    ddo['service'][0]['attributes']['main']['name'] = str(uuid.uuid4())
    return ddo


@pytest.fixture
def metadata():
    metadata = get_metadata()
    metadata['main']['files'][0]['checksum'] = str(uuid.uuid4())
    return metadata


@pytest.fixture
def algorithm_ddo():
    ddo = get_algorithm_ddo()
    ddo['service'][0]['attributes']['main']['checksum'] = str(uuid.uuid4())
    return ddo


@pytest.fixture
def workflow_ddo():
    ddo = get_workflow_ddo()
    ddo['service'][0]['attributes']['main']['checksum'] = str(uuid.uuid4())
    return ddo


@pytest.fixture
def setup_agreements_environment(ddo_sample):
    consumer_acc = get_consumer_account()
    publisher_acc = get_publisher_account()
    keeper = Keeper.get_instance()

    ddo = ddo_sample
    did_seed = generate_prefixed_id()
    asset_id = keeper.did_registry.hash_did(did_seed, publisher_acc.address)
    ddo._did = DID.did(asset_id)

    keeper.did_registry.register(
        did_seed,
        checksum=Web3Provider.get_web3().toBytes(hexstr=ddo.asset_id),
        url='localhost:5000',
        account=publisher_acc,
        providers=None
    )

    service_agreement = ServiceAgreement.from_ddo(ServiceTypes.ASSET_ACCESS, ddo)
    agreement_id = ServiceAgreement.create_new_agreement_id()
    price = service_agreement.get_price()
    access_cond_id, lock_cond_id, escrow_cond_id = \
        service_agreement.generate_agreement_condition_ids(
            agreement_id, asset_id, consumer_acc.address, keeper
        )

    return (
        keeper,
        publisher_acc,
        consumer_acc,
        agreement_id,
        asset_id,
        price,
        service_agreement,
        (lock_cond_id, access_cond_id, escrow_cond_id),
    )

@pytest.fixture
def setup_agreements_proof_environment(proof_ddo):
    consumer_acc = get_consumer_account()
    publisher_acc = get_publisher_account()
    consumer_acc.babyjub_address = get_buyer_public_key()
    publisher_acc.babyjub_address = get_provider_public_key()
    keeper = Keeper.get_instance()

    ddo = proof_ddo
    did_seed = generate_prefixed_id()
    asset_id = keeper.did_registry.hash_did(did_seed, publisher_acc.address)
    ddo._did = DID.did(asset_id)

    keeper.did_registry.register(
        did_seed,
        checksum=Web3Provider.get_web3().toBytes(hexstr=ddo.asset_id),
        url='localhost:5000',
        account=publisher_acc,
        providers=None
    )

    service_agreement = ServiceAgreement.from_ddo(ServiceTypes.ASSET_ACCESS_PROOF, ddo)
    agreement_id = ServiceAgreement.create_new_agreement_id()
    price = service_agreement.get_price()
    access_cond_id, lock_cond_id, escrow_cond_id = \
        service_agreement.generate_agreement_condition_ids(
            agreement_id, asset_id, consumer_acc.babyjub_address, keeper
        )

    return (
        keeper,
        publisher_acc,
        consumer_acc,
        agreement_id,
        asset_id,
        price,
        service_agreement,
        (lock_cond_id, access_cond_id, escrow_cond_id),
    )


@pytest.fixture
def agreements():
    publisher_acc = get_publisher_account()
    keeper = Keeper.get_instance()
    w3 = Web3Provider.get_web3()
    did_resolver = Mock()
    ddo = get_ddo_sample()
    service = ddo.get_service(ServiceTypes.ASSET_ACCESS)
    service.update_value(
        ServiceAgreementTemplate.TEMPLATE_ID_KEY,
        w3.toChecksumAddress(publisher_acc.address)
    )
    did_resolver.resolve = MagicMock(return_value=ddo)

    return Agreements(
        keeper,
        did_resolver,
        AssetConsumer,
        AssetExecutor,
        ConfigProvider.get_config()
    )



from nevermined_sdk_py import ConfigProvider
from nevermined_sdk_py.gateway.gateway import Gateway
from requests import Response

config = ConfigProvider.get_config()


def test_get_rsa_key():
    assert Gateway.get_rsa_public_key(config)


def test_get_file_name():
    response = Response()

    response.headers = {"content-disposition": "attachment;filename=test.txt"}
    print(response.headers.get("content-disposition"))
    file_name = Gateway._get_file_name(response)
    assert file_name == "test.txt"

    response.headers = {
        "content-disposition": "attachment;filename=test.txt?X-Amz-Algorithm=AWS4-HMAC-SHA256"
    }
    file_name = Gateway._get_file_name(response)
    assert file_name == "test.txt"



import pytest

def test_accounts(publisher_instance):
    for account in publisher_instance.accounts.list():
        print(account)

    # These accounts have a positive ETH balance
    for account in publisher_instance.accounts.list():
        assert publisher_instance.accounts.balance(account).eth >= 0
        assert publisher_instance.accounts.balance(account).ocn >= 0


def test_token_request(publisher_instance):
    receiver_account = publisher_instance.main_account
    # Starting balance for comparison
    start_ocean = publisher_instance.accounts.balance(receiver_account).ocn

    # Make requests, assert success on request
    publisher_instance.accounts.request_tokens(receiver_account, 2000)
    # Should be no change, 2000 exceeds the max of 1000
    assert publisher_instance.accounts.balance(receiver_account).ocn == start_ocean

    amount = 500
    publisher_instance.accounts.request_tokens(receiver_account, amount)
    # Confirm balance changes
    # TODO Review representation of amounts.
    assert publisher_instance.accounts.balance(
        receiver_account).ocn == start_ocean + (amount * 1000000000000000000)


@pytest.mark.skip(reason="It is only possible request once per 24hours")
def test_request_eth_from_faucet(consumer_instance):
    receiver_account = consumer_instance.main_account

    assert consumer_instance.accounts.request_eth_from_faucet(receiver_account.address)



from contracts_lib_py.account import Account

from nevermined_sdk_py.secret_store import SecretStoreProvider


class SSMock:
    def __init__(self, ss_url, keeper_url, account):
        self.ss_url = ss_url
        self.keeper_url = keeper_url
        self.account = account


def test_secret_store_provider():
    account = Account('0x00000000000000000', 'psw')
    SecretStoreProvider.set_secret_store_class(SSMock)
    ss = SecretStoreProvider.get_secret_store('ss/url', 'keeper/url', account)
    sss = SecretStoreProvider.get_secret_store('ss/url', 'keeper/url', account)
    assert ss != sss, ''
    assert isinstance(ss, SSMock)
    assert ss.ss_url == 'ss/url', ''
    assert ss.keeper_url == 'keeper/url', ''
    assert ss.account == account, ''



import logging

from contracts_lib_py.account import Account

from nevermined_sdk_py.secret_store.secret_store_provider import SecretStoreProvider

logger = logging.getLogger(__name__)


class SecretStore:
    """Nevermined SecretStore class."""

    def __init__(self, config):
        self._config = config
        if config.secret_store_url and config.parity_url and config.parity_address:
            logger.debug(f'\tSecretStore: url {config.secret_store_url}, '
                         f'parity-client {config.parity_url}, '
                         f'account {config.parity_address}')

        self._secret_store_url = config.secret_store_url
        self._parity_url = config.parity_url
        self._account = Account(config.parity_address, config.parity_password)

    def _secret_store(self, account):
        return SecretStoreProvider.get_secret_store(
            self._secret_store_url,
            self._parity_url,
            account or self._account
        )

    def encrypt(self, document_id, content, account):
        """
        Encrypt string data using the DID as an secret store id,
        if secret store is enabled then return the result from secret store encryption

        None for no encryption performed

        :param document_id: hex str id of document to use for encryption session
        :param content: str to be encrypted
        :param account: Account instance encrypting this content
        :return: hex str encrypted content
        """
        return self._secret_store(account).encrypt_document(document_id, content)

    def decrypt(self, document_id, encrypted_content, account):
        """
        Decrypt a previously encrypted content using the secret store keys identified
        by document_id.

        Note that decryption requires permission already granted to the consumer account.

        :param document_id: hex str id of document to use to retrieve the decryption keys
        :param encrypted_content: hex str
        :param account: Account instance to use for decrypting the `encrypted_content`
        :return: str original content
        """
        return self._secret_store(account).decrypt_document(document_id, encrypted_content)



def test_create_access_service(publisher_instance):
    service = publisher_instance.services.create_access_service(1, 'service_endpoint')
    assert service[0] == 'access'
    assert service[1]['attributes'] == 1
    assert service[1]['serviceEndpoint'] == 'service_endpoint'



import uuid


def new_did():
    return uuid.uuid4().hex + uuid.uuid4().hex


def test_get_provenance(publisher_instance, ddo_sample):
    publisher = publisher_instance.main_account
    activity_id = new_did()
    provenance_id = new_did()

    if publisher_instance.accounts.balance(publisher).ocn == 0:
        publisher_instance.accounts.request_tokens(publisher, 200)

    assert publisher_instance.accounts.balance(publisher).ocn > 0

    ddo = publisher_instance.assets.create(ddo_sample.metadata, publisher)

    assert publisher_instance.provenance.get_provenance_owner(ddo.did) == publisher.address

    publisher_instance.provenance.used(provenance_id, ddo.did, publisher.address, activity_id, "",
                                       account=publisher, attributes="used test")

    assert len(publisher_instance.provenance.get_did_provenance_events(ddo.did)) == 2

    provenance_entry = publisher_instance.provenance.get_provenance_entry(provenance_id)

    assert provenance_entry['activity_id'] == '0x' + activity_id
    assert provenance_entry['did'] == ddo.did
    assert provenance_entry['method'] == 3
    assert provenance_entry['created_by'] == publisher.address

    publisher_instance.assets.retire(ddo.did)


def test_delegate_provenance(publisher_instance, consumer_instance, ddo_sample):
    publisher = publisher_instance.main_account
    delegated = consumer_instance.main_account

    if publisher_instance.accounts.balance(publisher).ocn == 0:
        publisher_instance.accounts.request_tokens(publisher, 200)

    assert publisher_instance.accounts.balance(publisher).ocn > 0

    ddo = publisher_instance.assets.create(ddo_sample.metadata, publisher)

    assert publisher_instance.provenance.is_provenance_delegate(ddo.did, delegated.address) is False
    assert publisher_instance.provenance.add_did_provenance_delegate(ddo.did, delegated.address,
                                                                     publisher) is True
    assert publisher_instance.provenance.is_provenance_delegate(ddo.did, delegated.address) is True
    assert publisher_instance.provenance.remove_did_provenance_delegate(ddo.did, delegated.address,
                                                                        publisher) is True
    assert publisher_instance.provenance.is_provenance_delegate(ddo.did, delegated.address) is False


def test_search_multiple_provenance_event_tests(publisher_instance, ddo_sample):
    publisher = publisher_instance.main_account
    activity_id = new_did()
    provenance_id = new_did()
    derived_did = new_did()

    if publisher_instance.accounts.balance(publisher).ocn == 0:
        publisher_instance.accounts.request_tokens(publisher, 200)

    assert publisher_instance.accounts.balance(publisher).ocn > 0

    ddo = publisher_instance.assets.create(ddo_sample.metadata, publisher)

    publisher_instance.provenance.used(provenance_id, ddo.did, publisher.address, activity_id, "",
                                       account=publisher, attributes="used test")

    publisher_instance.provenance.was_derived_from(new_did(), derived_did, ddo.did,
                                                   publisher.address,
                                                   activity_id,
                                                   account=publisher, attributes="was derived from")

    publisher_instance.provenance.was_associated_with(new_did(), ddo.did, publisher.address,
                                                      activity_id,
                                                      account=publisher,
                                                      attributes="was associated with")

    publisher_instance.provenance.acted_on_behalf(new_did(), ddo.did, publisher.address,
                                                  publisher.address, activity_id, '',
                                                  account=publisher, attributes="acted on behalf")

    assert len(publisher_instance.provenance.get_did_provenance_methods_events('WAS_GENERATED_BY',
                                                                               ddo.did)) == 1
    assert len(
        publisher_instance.provenance.get_did_provenance_methods_events('USED', ddo.did)) == 1
    assert len(publisher_instance.provenance.get_did_provenance_methods_events('WAS_DERIVED_FROM',
                                                                               derived_did)) == 1
    assert len(
        publisher_instance.provenance.get_did_provenance_methods_events('WAS_ASSOCIATED_WITH',
                                                                        ddo.did)) == 1

    assert len(publisher_instance.provenance.get_did_provenance_methods_events('ACTED_ON_BEHALF',
                                                                               ddo.did)) == 1


def test_calling_twice_provenance(publisher_instance, ddo_sample):
    publisher = publisher_instance.main_account
    activity_id = new_did()
    provenance_id = new_did()

    if publisher_instance.accounts.balance(publisher).ocn == 0:
        publisher_instance.accounts.request_tokens(publisher, 200)

    assert publisher_instance.accounts.balance(publisher).ocn > 0

    ddo = publisher_instance.assets.create(ddo_sample.metadata, publisher)

    assert publisher_instance.provenance.used(provenance_id, ddo.did, publisher.address,
                                              activity_id, "",
                                              account=publisher, attributes="used test") is True

    assert publisher_instance.provenance.used(provenance_id, ddo.did, publisher.address,
                                              activity_id, "",
                                              account=publisher, attributes="used test") is False



import logging

logger = logging.getLogger(__name__)


class Templates:
    """Nevermined templates class."""

    def __init__(self, keeper, config):
        self._keeper = keeper
        self._config = config
        self.access_template_id = self._keeper.access_template.address

    def propose(self, template_address, account):
        """
        Propose a new template.

        :param template_address: Address of the template contract, str
        :param account: account proposing the template, Account
        :return: bool
        """
        try:
            proposed = self._keeper.template_manager.propose_template(template_address, account)
            return proposed
        except ValueError as err:
            template_values = self._keeper.template_manager.get_template(template_address)
            if not template_values:
                logger.warning(f'Propose template failed: {err}')
                return False

            if template_values.state != 1:
                logger.warning(
                    f'Propose template failed, current state is set to {template_values.state}')
                return False

            return True

    def approve(self, template_address, account):
        """
        Approve a template already proposed. The account needs to be owner of the templateManager
        contract to be able of approve the template.

        :param template_address: Address of the template contract, str
        :param account: account approving the template, Account
        :return: bool
        """
        try:
            approved = self._keeper.template_manager.approve_template(template_address, account)
            return approved
        except ValueError as err:
            template_values = self._keeper.template_manager.get_template(template_address)
            if not template_values:
                logger.warning(f'Approve template failed: {err}')
                return False

            if template_values.state == 1:
                logger.warning(f'Approve template failed, this template is '
                               f'currently in "proposed" state.')
                return False

            if template_values.state == 3:
                logger.warning(f'Approve template failed, this template appears to be '
                               f'revoked.')
                return False

            if template_values.state == 2:
                return True

            return False

    def revoke(self, template_address, account):
        """
        Revoke a template already approved. The account needs to be owner of the templateManager
        contract to be able of revoke the template.

        :param template_address: Address of the template contract, str
        :param account: account revoking the template, Account
        :return: bool
        """
        try:
            revoked = self._keeper.template_manager.revoke_template(template_address, account)
            return revoked
        except ValueError as err:
            template_values = self._keeper.template_manager.get_template(template_address)
            if not template_values:
                logger.warning(f'Cannot revoke template since it does not exist: {err}')
                return False

            logger.warning(f'Only template admin or owner can revoke a template: {err}')
            return False

    def get_list_size(self):
        """
        Return the number of templates approved.
        """
        return self._keeper.template_manager.get_num_templates()

    def get_template(self, template_id):
        """
        Get the template for a given template id.

        :param template_id: id of the template, str
        """
        return self._keeper.template_manager.get_template(template_id)

    def is_template_approved(self, template_address):
        """
        Return true is the template is approved
        :param template_address: Address of the template contract, str
        """
        return self._keeper.template_manager.is_template_approved(template_address)


import pytest

def test_get_service_endpoint(metadata_provider_instance):
    assert metadata_provider_instance.get_service_endpoint() == f'{metadata_provider_instance.url}/' + '{did}'


def test_publish_valid_ddo(asset1, metadata_provider_instance):
    metadata_provider_instance.publish_asset_ddo(asset1)
    assert metadata_provider_instance.get_asset_ddo(asset1.did)
    metadata_provider_instance.retire_asset_ddo(asset1.did)


def test_publish_invalid_ddo(metadata_provider_instance):
    with pytest.raises(AttributeError):
        metadata_provider_instance.publish_asset_ddo({})


def test_publish_ddo_already_registered(asset1, metadata_provider_instance):
    metadata_provider_instance.publish_asset_ddo(asset1)
    with pytest.raises(ValueError):
        metadata_provider_instance.publish_asset_ddo(asset1)
    metadata_provider_instance.retire_asset_ddo(asset1.did)


def test_get_asset_ddo_for_not_registered_did(metadata_provider_instance):
    invalid_did = 'did:nv:not_valid'
    assert metadata_provider_instance.get_asset_ddo(invalid_did) == {}


def test_get_asset_metadata(asset1, metadata_provider_instance):
    metadata_provider_instance.publish_asset_ddo(asset1)
    metadata_dict = metadata_provider_instance.get_asset_metadata(asset1.did)
    assert isinstance(metadata_dict, dict)
    assert 'main' in metadata_dict
    assert 'curation' in metadata_dict
    assert 'additionalInformation' in metadata_dict
    metadata_provider_instance.retire_asset_ddo(asset1.did)


def test_get_asset_metadata_for_not_registered_did(metadata_provider_instance):
    invalid_did = 'did:nv:not_valid'
    with pytest.raises(ValueError):
        metadata_provider_instance.get_asset_metadata(invalid_did)


def test_list_assets(asset1, metadata_provider_instance):
    num_assets = len(metadata_provider_instance.list_assets())
    metadata_provider_instance.publish_asset_ddo(asset1)
    assert len(metadata_provider_instance.list_assets()) == (num_assets + 1)
    assets = metadata_provider_instance.list_assets()
    assert isinstance(assets, list)
    assert isinstance(assets[0], str)
    metadata_provider_instance.retire_asset_ddo(asset1.did)


def test_list_assets_ddo(asset1, metadata_provider_instance):
    num_assets = len(metadata_provider_instance.list_assets_ddo())
    metadata_provider_instance.publish_asset_ddo(asset1)
    assert len(metadata_provider_instance.list_assets_ddo()) == (num_assets + 1)
    assert isinstance(metadata_provider_instance.list_assets_ddo(), dict)
    metadata_provider_instance.retire_asset_ddo(asset1.did)


def test_update_ddo(asset1, asset2, metadata_provider_instance):
    metadata_provider_instance.publish_asset_ddo(asset1)
    metadata_provider_instance.update_asset_ddo(asset1.did, asset2)
    assert metadata_provider_instance.get_asset_ddo(asset1.did).did == asset2.did
    assert metadata_provider_instance.get_asset_ddo(asset1.did).metadata['main']['name'] != \
           asset1.metadata['main'][
               'name'], 'The name has not been updated correctly.'
    metadata_provider_instance.retire_asset_ddo(asset1.did)


def test_update_with_not_valid_ddo(asset1, metadata_provider_instance):
    with pytest.raises(Exception):
        metadata_provider_instance.update_asset_ddo(asset1.did, {})


def test_text_search(asset1, asset2, metadata_provider_instance):
    office_matches = 0
    metadata_provider_instance.publish_asset_ddo(asset1)
    assert len(
        metadata_provider_instance.text_search(text='Weather information', offset=10000)['results']) == (
                   office_matches + 1)

    text = 'UK'
    id_matches2 = len(metadata_provider_instance.text_search(text=text, offset=10000)['results'])
    metadata_provider_instance.publish_asset_ddo(asset2)
    assert len(metadata_provider_instance.text_search(text=text, offset=10000)['results']) == (
            id_matches2 + 1)

    assert len(
        metadata_provider_instance.text_search(text='Weather information', offset=10000)['results']) == (
                   office_matches + 2)
    metadata_provider_instance.retire_asset_ddo(asset1.did)
    metadata_provider_instance.retire_asset_ddo(asset2.did)


def test_text_search_invalid_query(metadata_provider_instance):
    with pytest.raises(Exception):
        metadata_provider_instance.text_search(text='', offset='Invalid')


def test_query_search(asset1, asset2, metadata_provider_instance):
    num_matches = 0
    metadata_provider_instance.publish_asset_ddo(asset1)

    search_query = {
        "query": {
            "bool": {
                "must": [{
                    "match": {"service.attributes.main.type": "dataset"}
                }]
            }
        }
    }
    assert len(metadata_provider_instance.query_search(search_query=search_query,
                                              offset=10000)['results']) == (
                   num_matches + 1)

    metadata_provider_instance.publish_asset_ddo(asset2)

    assert len(metadata_provider_instance.query_search(search_query=search_query,
                                              offset=10000)['results']) == (
                   num_matches + 2)
    metadata_provider_instance.retire_asset_ddo(asset1.did)
    metadata_provider_instance.retire_asset_ddo(asset2.did)


def test_query_search_invalid_query(metadata_provider_instance):
    with pytest.raises(Exception):
        metadata_provider_instance.query_search(search_query='')


def test_retire_ddo(asset1, metadata_provider_instance):
    n = len(metadata_provider_instance.list_assets())
    metadata_provider_instance.publish_asset_ddo(asset1)
    assert len(metadata_provider_instance.list_assets()) == (n + 1)
    metadata_provider_instance.retire_asset_ddo(asset1.did)
    assert len(metadata_provider_instance.list_assets()) == n


def test_retire_all_ddos(asset1, metadata_provider_instance):
    metadata_provider_instance.retire_all_assets()
    assert len(metadata_provider_instance.list_assets()) == 0


def test_retire_not_published_did(metadata_provider_instance):
    with pytest.raises(Exception):
        metadata_provider_instance.retire_asset_ddo('did:nv:not_registered')



import logging

from contracts_lib_py import Keeper


class NeverminedKeeper(Keeper):

    @staticmethod
    def get_instance(artifacts_path=None, contract_names=None):
        return NeverminedKeeper(artifacts_path)

    def get_condition_name_by_address(self, address):
        """Return the condition name for a given address."""
        if self.lock_payment_condition.address == address:
            return 'lockReward'
        elif self.access_condition.address == address:
            return 'accessSecretStore'
        elif self.access_proof_condition.address == address:
            return 'accessProof'
        elif self.compute_execution_condition.address == address:
            return 'execCompute'
        elif self.escrow_payment_condition.address == address:
            return 'escrowReward'
        else:
            logging.error(f'The current address {address} is not a condition address')



from common_utils_py.utils.keytransfer import make_prover, prove_transfer, hash_key
from web3 import Web3

prover = make_prover("/usr/local/share/keytransfer/keytransfer.zkey", "/usr/local/share/keytransfer/keytransfer.dat")

def call_prover(consumer_pub, provider_secret, asset_plain):
    c = Web3.keccak(text=provider_secret)
    provider_key=c.hex()[0:60]
    res = prove_transfer(prover, [int(consumer_pub[0], 16), int(consumer_pub[1], 16)], int(provider_key, 16), bytes.fromhex(asset_plain[2:]))
    return res

def poseidon_hash(asset_plain):
    res = hash_key(bytes.fromhex(asset_plain))
    return res



from common_utils_py.did import did_to_id
from common_utils_py.utils.utilities import to_checksum_addresses
from eth_utils import add_0x_prefix


class Conditions:
    """Nevermined conditions class."""

    def __init__(self, keeper):
        self._keeper = keeper

    def lock_payment(self, agreement_id, asset_id, amounts, receivers, token_address, account):
        """
        Lock reward condition.

        :param agreement_id: id of the agreement, hex str
        :param asset_id: asset identifier, str
        :param amounts: Amount of tokens, str[]
        :param receivers: Tokens receivers, str[]
        :param token_address: ERC20 token address or 0x0 if ETH
        :param account: Account
        :return: bool
        """
        total_price = 0
        for amount in amounts:
            total_price = total_price + int(amount)
        self._keeper.token.token_approve(self._keeper.lock_payment_condition.address, total_price,
                                         account)

        tx_hash = self._keeper.lock_payment_condition.fulfill(
            agreement_id,
            asset_id,
            self._keeper.escrow_payment_condition.address,
            token_address,
            amounts,
            to_checksum_addresses(receivers),
            account)

        receipt = self._keeper.lock_payment_condition.get_tx_receipt(tx_hash)
        return bool(receipt and receipt.status == 1)

    def grant_access(self, agreement_id, asset_id, grantee_address, account):
        """
        Grant access condition.

        :param agreement_id: id of the agreement, hex str
        :param asset_id: asset id, str
        :param grantee_address: Address, hex str
        :param account: Account
        :return:
        """
        tx_hash = self._keeper.access_condition.fulfill(
            agreement_id, asset_id, grantee_address, account
        )
        receipt = self._keeper.access_condition.get_tx_receipt(tx_hash)
        return bool(receipt and receipt.status == 1)

    def release_reward(self, agreement_id, asset_id, amounts, receivers, token_address, account):
        """
        Release reward condition.

        :param agreement_id: id of the agreement, hex str
        :param asset_id: asset id, str
        :param amounts: Amounts of tokens, int[]
        :param receivers: Token receivers, str[]
        :param token_address: ERC20 address or 0x0 if ETH payment
        :param account sending the transaction
        :return:
        """
        agreement_values = self._keeper.agreement_manager.get_agreement(agreement_id)
        consumer, provider = self._keeper.access_template.get_agreement_data(
            agreement_id)
        owner = agreement_values.owner
        access_id, lock_id = agreement_values.condition_ids[:2]
        tx_hash = self._keeper.escrow_payment_condition.fulfill(
            agreement_id,
            asset_id,
            amounts,
            to_checksum_addresses(receivers),
            self._keeper.escrow_payment_condition.address,
            token_address,
            lock_id,
            access_id,
            account
        )
        receipt = self._keeper.escrow_payment_condition.get_tx_receipt(tx_hash)
        return bool(receipt and receipt.status == 1)

    def refund_reward(self, agreement_id, asset_id, amounts, receivers, account):
        """
        Refund reward condition.

        :param agreement_id: id of the agreement, hex str
        :param asset_id: asset id, str
        :param amounts: Amounts of tokens, int[]
        :param receivers: Token receivers, str[]
        :param account sending the transaction
        :return:
        """
        return self.release_reward(agreement_id, asset_id, amounts, receivers, account)

    def show_nft(self, agreement_id, asset_id, holder_address, number_nfts, account):
        """
        Fulfill the NFT Holder condition demonstrating that holder_address has x number of nfts
        for a specific asset_id

        :param agreement_id: id of the agreement, hex str
        :param asset_id: asset id, str
        :param holder_address: Address, hex str
        :param number_nfts: int
        :param account: Account
        :return:
        """
        tx_hash = self._keeper.nft_holder_condition.fulfill(
            agreement_id, asset_id, holder_address, number_nfts, account
        )
        receipt = self._keeper.nft_access_condition.get_tx_receipt(tx_hash)
        return bool(receipt and receipt.status == 1)

    def grant_nft_access(self, agreement_id, asset_id, grantee_address, account):
        """
        Grant nft access condition.

        :param agreement_id: id of the agreement, hex str
        :param asset_id: asset id, str
        :param grantee_address: Address, hex str
        :param account: Account
        :return:
        """
        tx_hash = self._keeper.nft_access_condition.fulfill(
            agreement_id, asset_id, grantee_address, account
        )
        receipt = self._keeper.nft_access_condition.get_tx_receipt(tx_hash)
        return bool(receipt and receipt.status == 1)


import json
import logging
import os
import re
import time
from common_utils_py.agreements.service_agreement import ServiceAgreement
from common_utils_py.exceptions import EncryptAssetUrlsError
from common_utils_py.http_requests.requests_session import get_requests_session
from common_utils_py.oauth2.token import (NeverminedJWTBearerGrant,
                                          generate_access_grant_token,
                                          generate_download_grant_token,
                                          generate_execute_grant_token,
                                          generate_compute_grant_token)
from contracts_lib_py.utils import add_ethereum_prefix_and_hash_msg
from nevermined_sdk_py.nevermined.keeper import NeverminedKeeper as Keeper

logger = logging.getLogger(__name__)


class Gateway:
    """
    `Gateway` is the name chosen for the asset service provider.

    The main functions available are:
    - Allow to get access to an asset data file
    - encrypt a secret using different methods
    - consume_service
    - run_compute_service

    """
    _http_client = get_requests_session()
    _tokens_cache = {}

    PING_ITERATIONS = 15
    PING_SLEEP = 1500

    @staticmethod
    def _generate_cache_key(*args):
        return ''.join(args)

    @staticmethod
    def set_http_client(http_client):
        """Set the http client to something other than the default `requests`"""
        Gateway._http_client = http_client

    @staticmethod
    def old_encrypt_files_dict(files_dict, encrypt_endpoint, asset_id, account_address, signed_did):
        payload = json.dumps({
            'documentId': asset_id,
            'signedDocumentId': signed_did,
            'document': json.dumps(files_dict),
            'publisherAddress': account_address
        })
        response = Gateway._http_client.post(
            encrypt_endpoint, data=payload,
            headers={'content-type': 'application/json'}
        )
        if response and hasattr(response, 'status_code'):
            if response.status_code != 201:
                msg = (f'Encrypt file urls failed at the encryptEndpoint '
                       f'{encrypt_endpoint}, reason {response.text}, status {response.status_code}'
                       )
                logger.error(msg)
                raise EncryptAssetUrlsError(msg)

            logger.info(
                f'Asset urls encrypted successfully, encrypted urls str: {response.text},'
                f' encryptedEndpoint {encrypt_endpoint}')

            return response.text

    @staticmethod
    def encrypt_files_dict(files_dict, encrypt_endpoint, asset_id, method):
        payload = json.dumps({
            'did': asset_id,
            'message': json.dumps(files_dict),
            'method': method
        })
        response = Gateway._http_client.post(
            encrypt_endpoint, data=payload,
            headers={'content-type': 'application/json'}
        )
        if response and hasattr(response, 'status_code'):
            if response.status_code != 200:
                msg = (f'Encrypt file urls failed at the encryptEndpoint '
                       f'{encrypt_endpoint}, reason {response.text}, status {response.status_code}'
                       )
                logger.error(msg)
                raise EncryptAssetUrlsError(msg)

            logger.info(
                f'Asset urls encrypted successfully, encrypted urls str: {response.text},'
                f' encryptedEndpoint {encrypt_endpoint}')

            return json.loads(response.text)['hash']

    @staticmethod
    def access_service(did, service_agreement_id, service_endpoint, account, destination_folder, config, index, uri='/access'):
        cache_key = Gateway._generate_cache_key(account.address, service_agreement_id, did)
        if cache_key not in Gateway._tokens_cache:
            grant_token = generate_access_grant_token(account, service_agreement_id, did, uri)
            access_token = Gateway.fetch_token(grant_token, config)
            Gateway._tokens_cache[cache_key] = access_token
        else:
            access_token = Gateway._tokens_cache[cache_key]

        consume_url = Gateway._create_access_url(service_endpoint, service_agreement_id, index)
        headers = {"Authorization": f"Bearer {access_token}"}

        response = Gateway._http_client.get(consume_url, headers=headers, stream=True)
        if not response.ok:
            raise ValueError(response.text)

        file_name = Gateway._get_file_name(response)
        Gateway.write_file(response, destination_folder, file_name or f'file-{index}')

        return response

    @staticmethod
    def compute_logs(service_agreement_id, execution_id, account, config):
        """Get the logs of a compute workflow.

        Args:
            service_agreement_id (str): The id of the service agreement.
            execution_id (str): The id of the workflow execution.
            account (:py:class:`contracts_lib_py.account.Account`): The account that ordered
                the execution of the workflow.
            config (:py:class:`nevermined_sdk_py.config.Config`): nevermined-sdk config instance.

        Returns:
            :py:class:`requests.Response`: HTTP server response

        """
        cache_key = Gateway._generate_cache_key(account.address, service_agreement_id, execution_id)
        if cache_key not in Gateway._tokens_cache:
            grant_token = generate_compute_grant_token(account, service_agreement_id, execution_id)
            access_token = Gateway.fetch_token(grant_token, config)
            Gateway._tokens_cache[cache_key] = access_token
        else:
            access_token = Gateway._tokens_cache[cache_key]

        headers = {"Authorization": f"Bearer {access_token}"}
        consume_url = Gateway._create_compute_logs_url(config, service_agreement_id, execution_id)

        iteration = 0
        while iteration < Gateway.PING_ITERATIONS:
            iteration = iteration + 1
            response = Gateway._http_client.get(consume_url, headers=headers)
            if not response.ok:
                time.sleep(Gateway.PING_SLEEP / 1000)
            else:
                return response
        raise ValueError(response.text)

    @staticmethod
    def compute_status(service_agreement_id, execution_id, account, config):
        """Get the status of a compute workflow.

        Args:
            service_agreement_id (str): The id of the service agreement.
            execution_id (str): The id of the workflow execution.
            account (:py:class:`contracts_lib_py.account.Account`): The account that ordered
                the execution of the workflow.
            config (:py:class:`nevermined_sdk_py.config.Config`): nevermined-sdk config instance.

        Returns:
            :py:class:`requests.Response`: HTTP server response

        """
        cache_key = Gateway._generate_cache_key(account.address, service_agreement_id, execution_id)
        if cache_key not in Gateway._tokens_cache:
            grant_token = generate_compute_grant_token(account, service_agreement_id, execution_id)
            access_token = Gateway.fetch_token(grant_token, config)
            Gateway._tokens_cache[cache_key] = access_token
        else:
            access_token = Gateway._tokens_cache[cache_key]

        headers = {"Authorization": f"Bearer {access_token}"}
        consume_url = Gateway._create_compute_status_url(config, service_agreement_id, execution_id)

        response = Gateway._http_client.get(consume_url, headers=headers)
        if response.status_code != 200:
            raise ValueError(response.text)
        return response

    @staticmethod
    def download(did, account, destination_folder, index, config):
        """Allows an asset data file if the account is the owner or provider of the asset

        Args:
            did (str): The did of the asset.
            account (:py:class:`contracts_lib_py.account.Account`): The account
                that should be the owner or provider of the asset.
            destination_folder (str): The destination where the downloaded file should be put.
            index (int): Index of the file.
            config (:py:class:`nevermined_sdk_py.config.Config`): nevermined-sdk config instance.

        Returns:
            :py:class:`requests.Response`: HTTP server response

        """
        cache_key = Gateway._generate_cache_key(account.address, did)
        if cache_key not in Gateway._tokens_cache:
            grant_token = generate_download_grant_token(account, did)
            access_token = Gateway.fetch_token(grant_token, config)
            Gateway._tokens_cache[cache_key] = access_token
        else:
            access_token = Gateway._tokens_cache[cache_key]

        headers = {"Authorization": f"Bearer {access_token}"}
        consume_url = Gateway._create_download_url(config, index)

        response = Gateway._http_client.get(consume_url, headers=headers, stream=True)
        if not response.ok:
            raise ValueError(response.text)

        file_name = Gateway._get_file_name(response)
        Gateway.write_file(response, destination_folder, file_name or f'file-{index}')

        return response

    @staticmethod
    def execute_service(service_agreement_id, service_endpoint, account, workflow_ddo):
        """

        :param service_agreement_id:
        :param service_endpoint:
        :param account:
        :return:
        """
        signature = Keeper.get_instance().sign_hash(
            add_ethereum_prefix_and_hash_msg(service_agreement_id),
            account)
        execute_url = Gateway._create_execute_url(service_endpoint, service_agreement_id, account,
                                                  workflow_ddo.did, signature)
        logger.info(f'invoke execute endpoint with this url: {execute_url}')
        response = Gateway._http_client.post(execute_url)
        return response

    @staticmethod
    def execute_compute_service(service_agreement_id, service_endpoint, account, workflow_ddo, config):
        cache_key = Gateway._generate_cache_key(account.address, service_agreement_id, workflow_ddo.did)
        if cache_key not in Gateway._tokens_cache:
            grant_token = generate_execute_grant_token(account, service_agreement_id, workflow_ddo.did)
            access_token = Gateway.fetch_token(grant_token, config)
            Gateway._tokens_cache[cache_key] = access_token
        else:
            access_token = Gateway._tokens_cache[cache_key]

        headers = {"Authorization": f"Bearer {access_token}"}
        execute_url = Gateway._create_compute_url(service_endpoint, service_agreement_id)

        response = Gateway._http_client.post(execute_url, headers=headers)
        if not response.ok:
            raise ValueError(response.text)
        return response

    @staticmethod
    def fetch_token(grant_token, config):
        fetch_token_url = Gateway.get_fetch_token_endpoint(config)
        response = Gateway._http_client.post(fetch_token_url, data={
            "grant_type": NeverminedJWTBearerGrant.GRANT_TYPE,
            "assertion": grant_token
        })
        if not response.ok:
            raise ValueError(response.text)
        return response.json()["access_token"]

    @staticmethod
    def upload_filecoin(file_, config):
        upload_filecoin_url = Gateway.get_upload_filecoin_endpoint(config)
        response = Gateway._http_client.post(upload_filecoin_url, files={
            'file': file_
        })

        if not response.ok:
            raise ValueError(response.text)

        return response.json()

    @staticmethod
    def _prepare_consume_payload(did, service_agreement_id, service_index, signature,
                                 consumer_address):
        """Prepare a payload to send to Nevermined Gateway.

        :param did: DID, str
        :param service_agreement_id: Service Agreement Id, str
        :param service_index: identifier of the service inside the asset DDO, str
        service in the DDO (DID document)
        :param signature: the signed agreement message hash which includes
         conditions and their parameters values and other details of the agreement, str
        :param consumer_address: ethereum address of the consumer signing this agreement, hex-str
        :return: dict
        """
        return json.dumps({
            'did': did,
            'serviceAgreementId': service_agreement_id,
            ServiceAgreement.SERVICE_INDEX: service_index,
            'signature': signature,
            'consumerAddress': consumer_address
        })

    @staticmethod
    def get_gateway_url(config):
        """
        Return the Gateway component url.

        :param config: Config
        :return: Url, str
        """
        gateway_url = 'http://localhost:8030'
        if config.has_option('resources', 'gateway.url'):
            gateway_url = config.get('resources', 'gateway.url') or gateway_url

        gateway_path = '/api/v1/gateway'
        return f'{gateway_url}{gateway_path}'

    @staticmethod
    def get_access_endpoint(config):
        """
        Return the endpoint to access the asset.

        :param config:Config
        :return: Url, str
        """
        return f'{Gateway.get_gateway_url(config)}/services/access'

    @staticmethod
    def get_access_proof_endpoint(config):
        """
        Return the endpoint to access the asset.

        :param config:Config
        :return: Url, str
        """
        return f'{Gateway.get_gateway_url(config)}/services/access-proof'

    @staticmethod
    def get_nft_access_endpoint(config):
        """
        Return the endpoint to access the asset.

        :param config:Config
        :return: Url, str
        """
        return f'{Gateway.get_gateway_url(config)}/services/nft-access'

    @staticmethod
    def get_download_endpoint(config):
        """
        Return he endpoint to download the asset.

        :param config:Config
        :return: Url, str
        """
        return f'{Gateway.get_gateway_url(config)}/services/download'

    @staticmethod
    def get_compute_logs_endpoint(config):
        """
        Return he endpoint to get the logs of a compute workflow.

        :param config:Config
        :return: Url, str
        """
        return f'{Gateway.get_gateway_url(config)}/services/compute/logs'

    @staticmethod
    def get_compute_status_endpoint(config):
        """
        Return he endpoint to get the status of a compute workflow.

        :param config:Config
        :return: Url, str
        """
        return f'{Gateway.get_gateway_url(config)}/services/compute/status'

    @staticmethod
    def get_consume_endpoint(config):
        """
        Return the url to access the asset.

        :param config: Config
        :return: Url, str
        """
        return f'{Gateway.get_gateway_url(config)}/services/access'

    @staticmethod
    def get_execute_endpoint(config):
        """
        Return the url to execute the asset.

        :param config: Config
        :return: Url, str
        """
        return f'{Gateway.get_gateway_url(config)}/services/execute'

    @staticmethod
    def get_encrypt_endpoint(config):
        """
        Return the url to encrypt the asset.

        :param config: Config
        :return: Url, str
        """
        return f'{Gateway.get_gateway_url(config)}/services/encrypt'

    @staticmethod
    def get_exec_endpoint(config):
        """
        Return the url to execute the asset.
        This method has been deprecated.

        :param config: Config
        :return: Url, str
        """
        return f'{Gateway.get_gateway_url(config)}/services/exec'

    @staticmethod
    def get_publish_endpoint(config):
        """
        Return the url to encrypt the asset.
        This method has been deprecated.

        :param config: Config
        :return: Url, str
        """
        return f'{Gateway.get_gateway_url(config)}/services/publish'

    @staticmethod
    def get_fetch_token_endpoint(config):
        """
        Return the url to fetch an access token.

        :param config: Config
        :return: Url, str
        """
        return f'{Gateway.get_gateway_url(config)}/services/oauth/token'

    @staticmethod
    def get_upload_filecoin_endpoint(config):
        """
        Return the url to upload a file to Filecoin.

        :param config: Config,
        :return Url, str
        """
        return f'{Gateway.get_gateway_url(config)}/services/upload/filecoin'

    @staticmethod
    def get_rsa_public_key(config):
        return Gateway._http_client.get(config.gateway_url).json()['rsa-public-key']

    @staticmethod
    def get_ecdsa_public_key(config):
        return Gateway._http_client.get(config.gateway_url).json()['ecdsa-public-key']

    @staticmethod
    def _get_file_name(response):
        try:
            return re.match(r'attachment;filename=(.+?)($|\?)',
                            response.headers.get('content-disposition'))[1]
        except Exception as e:
            logger.warning(f'It was not possible to get the file name. {e}')

    @staticmethod
    def write_file(response, destination_folder, file_name):
        """
        Write the response content in a file in the destination folder.
        :param response: Response
        :param destination_folder: Destination folder, string
        :param file_name: File name, string
        :return: bool
        """
        if response.status_code == 200:
            with open(os.path.join(destination_folder, file_name), 'wb') as f:
                for chunk in response.iter_content(chunk_size=None):
                    f.write(chunk)
            logger.info(f'Saved downloaded file in {f.name}')
        else:
            logger.warning(f'access failed: {response.reason}')

    @staticmethod
    def _create_access_url(service_endpoint, service_agreement_id, index=None):
        """Return the url to get access to an asset."""
        return f'{service_endpoint}/{service_agreement_id}/{index}'

    @staticmethod
    def _create_download_url(config, index=None):
        """Return the url to download an asset."""
        return f'{Gateway.get_download_endpoint(config)}/{index}'

    @staticmethod
    def _create_compute_logs_url(config, service_agreement_id, execution_id):
        """Return the url to get the execution logs of a compute workflow."""
        return f'{Gateway.get_compute_logs_endpoint(config)}/{service_agreement_id}/{execution_id}'

    @staticmethod
    def _create_compute_status_url(config, service_agreement_id, execution_id):
        """Return the url to get the status of a compute workflow."""
        return f'{Gateway.get_compute_status_endpoint(config)}/{service_agreement_id}/{execution_id}'

    @staticmethod
    def _create_consume_url(service_endpoint, service_agreement_id, account, _file=None,
                            signature=None, index=None):
        if _file is not None and 'url' in _file:
            url = _file['url']
            if url.startswith('"') or url.startswith("'"):
                url = url[1:-1]
            return (f'{service_endpoint}'
                    f'?url={url}'
                    f'&serviceAgreementId={service_agreement_id}'
                    f'&consumerAddress={account.address}'
                    )
        else:
            return (f'{service_endpoint}'
                    f'?signature={signature}'
                    f'&serviceAgreementId={service_agreement_id}'
                    f'&consumerAddress={account.address}'
                    f'&index={index}'
                    )

    @staticmethod
    def _create_execute_url(service_endpoint, service_agreement_id, account, workflow_did,
                            signature=None):
        return (f'{service_endpoint}'
                f'?signature={signature}'
                f'&serviceAgreementId={service_agreement_id}'
                f'&consumerAddress={account.address}'
                f'&workflowDID={workflow_did}')

    @staticmethod
    def _create_compute_url(service_endpoint, service_agreement_id):
        return f'{service_endpoint}/{service_agreement_id}'


import logging

from common_utils_py.http_requests.requests_session import get_requests_session
from contracts_lib_py.web3_provider import Web3Provider


logger = logging.getLogger(__name__)


class Faucet:
    """
    `Faucet`.

    The main functions available are:
    - Request eht for an account

    """
    _http_client = get_requests_session()

    @staticmethod
    def set_http_client(http_client):
        """Set the http client to something other than the default `requests`"""
        Faucet._http_client = http_client

    @staticmethod
    def get_faucet_url(config):
        """
        Return the Gateway component url.

        :param config: Config
        :return: Url, str
        """
        faucet_url = 'http://localhost:3001'
        if config.has_option('resources', 'faucet.url'):
            faucet_url = config.get('resources', 'faucet.url') or faucet_url

        faucet_path = '/faucet'
        return f'{faucet_url}{faucet_path}'

    @staticmethod
    def get_eth_from_faucet(config, address):
        payload = {'address': Web3Provider.get_web3().toChecksumAddress(address), 'agent': 'sdk-py'}
        response = Faucet._http_client.post(Faucet.get_faucet_url(config), data=payload, headers={'Accept': 'application/json'})

        if not response.ok:
            raise ValueError(response.text)

        return response




from nevermined_sdk_py.gateway.gateway import Gateway
import pathlib

from nevermined_sdk_py.gateway.gateway_provider import GatewayProvider


class Files:
    """Nevermined Files class."""

    def __init__(self, config):
        self._config = config

    def upload_filecoin(self, file_path):
        """Upload a file to Filecoin

        param file_path: The path to the file to upload, str
        return: The url with the cid, str
        """
        path = pathlib.Path(file_path)
        if not path.is_file():
            raise ValueError(f'File {file_path} not found or not a file')

        with path.open('rb') as f:
            gateway = GatewayProvider.get_gateway()
            response =  gateway.upload_filecoin(f, self._config)
            return response['url']


import logging
from datetime import datetime

from contracts_lib_py.utils import add_ethereum_prefix_and_hash_msg
from contracts_lib_py.web3_provider import Web3Provider

from nevermined_sdk_py import ConfigProvider
from nevermined_sdk_py.data_store.auth_tokens import AuthTokensStorage


class Auth:
    """Nevermined auth class.
    Provide basic management of a user auth token. This token can be used to emulate
    sign-in behaviour. The token can be stored and associated with an expiry time.
    This is useful in front-end applications that interact with a 3rd-party wallet
    apps. The advantage of using the auth token is to reduce the number of confirmation
    prompts requiring user action.

    The auth token works with a provider service such as Gateway which also uses this
    ocean module to handle auth tokens.

    Token format is "signature-timestamp".

    """
    DEFAULT_EXPIRATION_TIME = 30 * 24 * 60 * 60  # in seconds
    DEFAULT_MESSAGE = "Nevermined Protocol Authentication"

    def __init__(self, keeper, storage_path):
        self._keeper = keeper
        self._tokens_storage = AuthTokensStorage(storage_path)

    @staticmethod
    def _get_timestamp():
        return int(datetime.now().timestamp())

    def _get_expiration(self):
        return int(ConfigProvider.get_config().auth_token_expiration
                   or self.DEFAULT_EXPIRATION_TIME)

    def _get_raw_message(self):
        return ConfigProvider.get_config().auth_token_message or self.DEFAULT_MESSAGE

    def _get_message(self, timestamp):
        return f'{self._get_raw_message()}\n{timestamp}'

    def _get_message_and_time(self):
        timestamp = self._get_timestamp()
        return self._get_message(timestamp), timestamp

    @staticmethod
    def is_token_valid(token):
        return isinstance(token, str) and token.startswith('0x') and len(token.split('-')) == 2

    def get(self, account):
        """
        :param account: Account instance signing the token
        :return: hex str the token generated/signed by account
        """
        _message, _time = self._get_message_and_time()
        msg_hash = Web3Provider.get_web3().keccak(text=_message)
        try:
            prefixed_msg_hash = self._keeper.sign_hash(
                add_ethereum_prefix_and_hash_msg(msg_hash), account)
            return f'{prefixed_msg_hash}-{_time}'
        except Exception as e:
            logging.error(f'Error signing token: {str(e)}')

    def check(self, token):
        """
        :param token: hex str consist of signature and timestamp
        :return: hex str ethereum address
        """
        parts = token.split('-')
        if len(parts) < 2:
            return '0x0'

        sig, timestamp = parts
        if self._get_timestamp() > (int(timestamp) + self._get_expiration()):
            return '0x0'

        message = self._get_message(timestamp)
        address = self._keeper.personal_ec_recover(Web3Provider.get_web3().keccak(text=message), sig)
        return Web3Provider.get_web3().toChecksumAddress(address)

    def store(self, account):
        """
        :param account: Account instance signing the token
        :return:
            token that was generated and stored for this account
        """
        token = self.get(account)
        timestamp = token.split('-')[1]
        self._tokens_storage.write_token(account.address, token, timestamp)
        return token

    def restore(self, account):
        """
        :param account: Account instance to fetch the saved token
        :return:
            hex str the token retreived from storage
            None if no token found for this account
        """
        token = self._tokens_storage.read_token(account.address)[0]
        if not token:
            return None

        address = self.check(token)

        return token if address == account.address else None

    def is_stored(self, account):
        """
        :param account: Account instance
        :return: bool whether this account has a stored token
        """
        return self.restore(account) is not None



from collections import namedtuple

from contracts_lib_py.account import Account

Balance = namedtuple('Balance', ('eth', 'ocn'))


class Accounts:
    """Nevermined accounts class."""

    def __init__(self, keeper, config, nevermined_tokens, faucet):
        self._keeper = keeper
        self._config = config
        self._nevermined_tokens = nevermined_tokens
        self._faucet = faucet
        self._accounts = [Account(account_address) for account_address in self._keeper.accounts]
        if config.parity_address and config.parity_password:
            address = config.parity_address.lower()
            for account in self._accounts:
                if account.address.lower() == address:
                    account.password = config.parity_password
                    break

    @property
    def _accounts_addresses(self):
        """
        Return a list with the account addresses.

        :return: list
        """
        return [a.address for a in self._accounts]

    def list(self):
        """
        Return list of Account instances available in the current ethereum node

        :return: list of Account instances
        """
        return self._accounts[:]

    def balance(self, account):
        """
        Return the balance, a tuple with the eth and nevermined balance.

        :param account: Account instance to return the balance of
        :return: Balance tuple of (eth, ocn)
        """
        return Balance(self._keeper.get_ether_balance(account.address),
                       self._keeper.token.get_token_balance(account.address))

    def request_tokens(self, account, amount):
        """
        Request an amount of ocean tokens for an account.

        :param account: Account instance making the tokens request
        :param amount: int amount of tokens requested
        :raises OceanInvalidTransaction: if transaction fails
        :return: bool
        """
        return self._nevermined_tokens.request(account, amount)

    def request_eth_from_faucet(self, address):
        """
        Requests Network ETH to the faucet for paying the transactions gas
        :param address: eth address requesting, hex str
        """
        return self._faucet.get_eth_from_faucet(self._config, address)



class SecretStoreMock:
    args = None
    id_to_document = dict()

    def __init__(self, secret_store_url, keeper_url, account):
        self.args = (secret_store_url, keeper_url, account)

    def set_secret_store_url(self, url):
        return

    def encrypt_document(self, document_id, document, threshold=0):
        encrypted = f'{self}.{document_id}!!{document}!!{threshold}'
        self.id_to_document[document_id] = (document, encrypted)
        return encrypted

    def decrypt_document(self, document_id, encrypted_document):
        doc, encrypted = self.id_to_document.get(document_id)
        assert encrypted == encrypted_document
        return doc



from nevermined_sdk_py.nevermined.keeper import NeverminedKeeper as Keeper


def test_get_condition_name_by_address():
    keeper = Keeper.get_instance()
    name = keeper.get_condition_name_by_address(keeper.lock_payment_condition.address)
    assert name == 'lockReward'

    name = keeper.get_condition_name_by_address(keeper.access_condition.address)
    assert name == 'accessSecretStore'

    name = keeper.get_condition_name_by_address(keeper.escrow_payment_condition.address)
    assert name == 'escrowReward'



import logging

from eth_utils import remove_0x_prefix
from secret_store_client.client import Client

logger = logging.getLogger(__name__)


class SecretStore(object):
    """Wrapper around the secret store client."""
    _client_class = Client

    def __init__(self, url, parity_url, account):
        """
        :param url: str URL of secret store node to create a document encryption session and store
            the decryption keys
        :param parity_url: str URL of parity/ethereum node to use for
            signing and encryption/decryption
        :param account: Account instance of ethereum account to either encrypt or decrypt a document
        """
        self._secret_store_url = url
        self._parity_node_url = parity_url
        self._account = account

    @staticmethod
    def set_client(secret_store_client):
        """Set a different secret store client.

        :param secret_store_client: SecretStore client
        """
        SecretStore._client_class = secret_store_client

    def _secret_store_client(self, account):
        return SecretStore._client_class(
            self._secret_store_url, self._parity_node_url, account.address, account.password
        )

    def set_secret_store_url(self, url):
        """
        Set secret store url.

        :param url: Url, str
        """
        self._secret_store_url = url

    def encrypt_document(self, document_id, content, threshold=0):
        """
        encrypt string data using the DID as an secret store id,
        if secret store is enabled then return the result from secret store encryption

        None for no encryption performed

        :param document_id: hex str id of document to use for encryption session
        :param content: str to be encrypted
        :param threshold: int
        :return:
            None -- if encryption failed
            hex str -- the encrypted document
        """
        return self._secret_store_client(self._account).publish_document(
            remove_0x_prefix(document_id), content, threshold
        )

    def decrypt_document(self, document_id, encrypted_content):
        """
        Decrypt a previously encrypted content using the secret store keys identified
        by document_id.

        Note that decryption requires permission already granted to the consumer account.

        :param document_id: hex str id of document to use for encryption session
        :param encrypted_content: hex str -- the encrypted content from a previous
            `encrypt_document` operation
        :return:
            None -- if decryption failed
            str -- the original content that was encrypted previously
        """
        return self._secret_store_client(self._account).decrypt_document(
            remove_0x_prefix(document_id),
            encrypted_content
        )



import logging

from common_utils_py.did import id_to_did, convert_to_bytes
from contracts_lib_py.web3_provider import Web3Provider
from eth_utils import add_0x_prefix


class Provenance:
    """Nevermined provenance class."""

    def __init__(self, keeper, config):
        self._keeper = keeper
        self._config = config

    def used(self, provenance_id, did, agent_id, activity_id, signature, account, attributes=None):
        """
        Implements the W3C PROV Usage action

        :param provenance_id: provenance ID
        :param did: Identifier of the entity created
        :param agent_id: Agent Identifier
        :param activity_id: Identifier of the activity creating the new entity
        :param signature: Signature (optional) provided by the agent involved
        :param attributes: Attributes associated with the action
        :param account: Account making the call

        :return: true if the provenance event was registered correctly

        """
        try:
            receipt = self._keeper.did_registry.used(
                convert_to_bytes(provenance_id),
                convert_to_bytes(did),
                convert_to_bytes(agent_id),
                convert_to_bytes(activity_id),
                signature, account, attributes)
            return bool(receipt and receipt.status == 1)
        except Exception as e:
            logging.critical(f'On-chain call error: {e}')
            return False

    def was_derived_from(self, provenance_id, new_entity_did, used_entity_did, agent_id,
                         activity_id, account, attributes=None):
        """
        Implements the W3C PROV Derivation action

        :param provenance_id: provenance ID
        :param new_entity_did: Identifier of the new entity derived
        :param used_entity_did: Identifier of the entity used to derive the new entity
        :param agent_id: Agent Identifier
        :param activity_id: Identifier of the activity creating the new entity
        :param attributes: Attributes associated with the action
        :param account: Account making the call

        :return: true if the provenance event was registered correctly
        """
        try:
            receipt = self._keeper.did_registry.was_derived_from(
                convert_to_bytes(provenance_id),
                convert_to_bytes(new_entity_did),
                convert_to_bytes(used_entity_did),
                convert_to_bytes(agent_id),
                convert_to_bytes(activity_id),
                account, attributes)
            return bool(receipt and receipt.status == 1)
        except Exception as e:
            logging.critical(f'On-chain call error: {e}')
        return False

    def was_associated_with(self, provenance_id, did, agent_id, activity_id, account, attributes):
        """
        Implements the W3C PROV Association action

        :param provenance_id: provenance ID
        :param did: Identifier of the entity created
        :param agent_id: Agent Identifier
        :param activity_id: Identifier of the activity creating the new entity
        :param attributes: Attributes associated with the action
        :param account: Account making the call

        :return: true if the provenance event was registered correctly
        """
        try:
            receipt = self._keeper.did_registry.was_associated_with(
                convert_to_bytes(provenance_id),
                convert_to_bytes(did),
                convert_to_bytes(agent_id),
                convert_to_bytes(activity_id),
                account,
                attributes)
            return bool(receipt and receipt.status == 1)
        except Exception as e:
            logging.critical(f'On-chain call error: {e}')
            return False

    def acted_on_behalf(self, provenance_id, did, delegate_agent_id, responsible_agent_id,
                        activity_id, signature, account, attributes):
        """
        Implements the W3C PROV Association action

        :param provenance_id: provenance ID
        :param did: Identifier of the entity created
        :param delegate_agent_id: Delegate Agent Identifier
        :param responsible_agent_id:  Responsible Agent Identifier
        :param activity_id: Identifier of the activity creating the new entity
        :param attributes: Attributes associated with the action
        :param signature: Signature provided by the agent involved
        :param account: Account making the call

        :return: true if the provenance event was registered correctly
        """
        try:

            receipt = self._keeper.did_registry.acted_on_behalf(convert_to_bytes(provenance_id),
                                                                convert_to_bytes(did),
                                                                convert_to_bytes(
                                                                    delegate_agent_id),
                                                                convert_to_bytes(
                                                                    responsible_agent_id),
                                                                convert_to_bytes(activity_id),
                                                                signature,
                                                                account, attributes)
            return bool(receipt and receipt.status == 1)
        except Exception as e:
            logging.critical(f'On-chain call error: {e}')
            return False

    def get_provenance_entry(self, provenance_id):
        """
        Fetch from the on-chain Provenance registry the information about one provenance event, given a provenance id

        :param provenance_id: provenance ID
        :return: information on-chain related with the provenance
        """
        provenance_entry = self._keeper.did_registry.get_provenance_entry(
            convert_to_bytes(provenance_id))
        provenance_entry['did'] = id_to_did(provenance_entry['did'])
        provenance_entry['related_did'] = id_to_did(provenance_entry['related_did'])
        return provenance_entry

    def is_provenance_delegate(self, did, delegate):
        """
        Indicates if an address is a provenance delegate for a given DID
        :param did: Identifier of the asset
        :param delegate: delegate address of the delegate
        :return: true if the address is a provenance delegate
        """
        return self._keeper.did_registry.is_provenance_delegate(convert_to_bytes(did), delegate)

    def add_did_provenance_delegate(self, did, delegated_address, account):
        """
        Adds an address as delegate for a given DID
        :param did: Identifier of the asset
        :param delegated_address: delegate address of the delegate
        :param account: Account making the call
        :return: true if the address is a provenance delegate
        """
        return self._keeper.did_registry.add_did_provenance_delegate(convert_to_bytes(did),
                                                                     delegated_address, account)

    def remove_did_provenance_delegate(self, did, delegated_address, account):
        """
        Remove an address as delegate for a given DID
        :param did: Identifier of the asset
        :param delegated_address: delegate address of the delegate
        :param account: Account making the call
        :return: true if the address is a provenance delegate
        """
        return self._keeper.did_registry.remove_did_provenance_delegate(convert_to_bytes(did),
                                                                        delegated_address, account)

    def get_provenance_owner(self, provenance_id):
        """
        Get the provenance owner.
        :param provenance_id: provenance ID

        :return: String with the address owning the provenance entry
        """
        return self._keeper.did_registry.get_provenance_owner(convert_to_bytes(provenance_id))

    def get_did_provenance_events(self, did):
        """
        Search for ProvenanceAttributeRegistered events related with a specific DID
        :param did: Identifier of the asset

        :return: list of provenance events.
        """
        return self._keeper.did_registry.get_did_provenance_events(
            convert_to_bytes(did))

    def get_did_provenance_methods_events(self, method, did):
        """
        Search for provenance methods (used, wasGeneratedBy, etc.) given a DID

        :param method: Provenance methods to fetch
        :param did: Identifier of the asset

        :return: list of provenance events.

        """
        return self._keeper.did_registry.get_provenance_method_events(method,
                                                                      convert_to_bytes(did))




import pytest

from examples import ExampleConfig
from tests.resources.helper_functions import (get_algorithm_ddo,
                                              get_consumer_instance,
                                              get_publisher_instance)


def test_publisher_download_compute(metadata):
    config = ExampleConfig.get_config()
    publisher_instance = get_publisher_instance(True, False, False)
    publisher = publisher_instance.main_account
    ddo = publisher_instance.assets.create_compute(metadata, publisher)
    assert ddo

    service_agreement = ddo.get_service("compute")
    assert publisher_instance.assets.download(
        ddo.did, service_agreement.index, publisher, config.downloads_path
    )
    publisher_instance.assets.retire(ddo.did)


def test_consumer_download_compute(metadata):
    config = ExampleConfig.get_config()
    publisher_instance = get_publisher_instance(True, False, False)
    publisher = publisher_instance.main_account
    consumer_instance = get_consumer_instance(True, False, False)
    consumer = consumer_instance.main_account

    # publisher creates compute asset
    ddo = publisher_instance.assets.create_compute(metadata, publisher)
    assert ddo

    # consumer orders compute asset
    service_agreement = ddo.get_service("compute")
    _agreement_id = consumer_instance.assets.order(
        ddo.did, service_agreement.index, consumer, consumer
    )

    # consumer tries to download the asset
    with pytest.raises(ValueError):
        consumer_instance.assets.download(
            ddo.did, service_agreement.index, consumer, config.downloads_path
        )
    publisher_instance.assets.retire(ddo.did)


def test_publisher_download_algorithm(ddo_algorithm):
    config = ExampleConfig.get_config()
    publisher_instance = get_publisher_instance(True, False, False)
    publisher = publisher_instance.main_account
    metadata = ddo_algorithm["service"][0]
    ddo = publisher_instance.assets.create(metadata["attributes"], publisher)
    assert ddo

    service_agreement = ddo.get_service("access")
    assert publisher_instance.assets.download(
        ddo.did, service_agreement.index, publisher, config.downloads_path
    )
    publisher_instance.assets.retire(ddo.did)


def test_provider_download_algorithm(ddo_algorithm):
    config = ExampleConfig.get_config()
    publisher_instance = get_publisher_instance(True, False, False)
    provider = publisher_instance.main_account
    consumer_instance = get_consumer_instance(True, False, False)
    consumer = consumer_instance.main_account
    metadata = ddo_algorithm["service"][0]

    # consumer creates algorithm asset (for a compute job)
    ddo = consumer_instance.assets.create(
        metadata["attributes"], consumer, providers=[provider.address]
    )
    assert ddo

    # provider downloads the asset
    service_agreement = ddo.get_service("access")
    assert publisher_instance.assets.download(
        ddo.did, service_agreement.index, consumer, config.downloads_path
    )
    consumer_instance.assets.retire(ddo.did)



from nevermined_sdk_py.secret_store.secret_store import SecretStore


class SecretStoreProvider(object):
    """Provides the SecretStore instance."""
    _secret_store_class = SecretStore

    @staticmethod
    def get_secret_store(secret_store_url, keeper_url, account):
        """ Get an SecretStore instance."""
        return SecretStoreProvider._secret_store_class(secret_store_url, keeper_url, account)

    @staticmethod
    def set_secret_store_class(secret_store_class):
        """
         Set a SecretStore class

        :param secret_store_class: SecretStore class
        :return:  New SecretStore instance.
        """
        SecretStoreProvider._secret_store_class = secret_store_class
